/* pattern.hh
* This file is part of Colorifer
* Copyright (c) 2002,2003,2004 by Stanislav Ievlev
*
* This file is covered by the GNU General Public License,
* which should be included with colorifer as the file COPYING.
*/

#ifndef COLORIFER_PATTERN_HH__
#define COLORIFER_PATTERN_HH__

#include <string>
#include <utility>
#include <map>
#include <vector>

#include <boost/shared_ptr.hpp>

#include <pcre_regex.hh>
#include <ncurses/manip.hh>


namespace colorifer
{
	typedef std::pair<ncurses::color_t, ncurses::attr_t> color_pair;
	typedef std::vector<color_pair>	pair_vector;
	typedef std::map<int, pair_vector> color_map;
	
	const ncurses::color_t end_color = -1;

	/* generic engine for parsing color definitions like 'color,[attr]'*/
	struct pattern
	{
		pattern();
		virtual ~pattern() {}
		virtual void fill(const std::string& line,color_map& subcolors) = 0;

		ncurses::color_t fg_; /**< color for the pattern */
		ncurses::attr_t attr_; /**< color for the pattern */
	protected:
		void parse_colors(const std::string&);
		static ing::regex	com_re_; /**< command matching regexp */
	};
	
	/* basic address match style command 'some-addr p color'*/
	struct addr_pattern: public pattern
	{
		addr_pattern(const std::string&);
		void fill(const std::string& line,color_map& subcolors);
	protected:
		static ing::regex	com_re_; /**< command matching regexp */
		bool		positive_; /*positive or negative checks */
	};

	/** 
	 * command with pattern addressing '/regex/'
	 * @todo add support for %regex%
	 * @todo add support for I modifier
	 */
	struct regex_pattern: public addr_pattern
	{
		regex_pattern(const std::string&);
		void fill(const std::string& line,color_map& subcolors);
	private:
		boost::shared_ptr<ing::regex>	re_;
		static ing::regex	com_re_; /**< command matching regexp */
	};
	
	
	/**
	 * command with pattern addressing NUM or FIRST~SEP or 
	 *  ADDR1,ADDR2 or ADDR1,+N or ADDR1,~N
	 */
	struct lnum_pattern: public addr_pattern
	{
		lnum_pattern(const std::string&);
		void fill(const std::string& line,color_map& subcolors);
	private:
		int counter_; /**< line counter */
		int start_; /**< begin line number for colorize */
		int stop_; /**< end line numbe for colorize*/
		int each_; /**< colorize each Nth string */

		static ing::regex	com_re_; /**< command matching regexp */
	};

	/**
	 * command for most powerfull color subsitutions 's/regex/color/flags'
	 * @todo add support for flags
	 */
	struct regsubst_pattern: public pattern
	{
		regsubst_pattern(const std::string&);
		void fill(const std::string& line,color_map& subcolors);
	private:
		boost::shared_ptr<ing::regex>	re_;
		void set_colors(int start, int stop, color_map& subcolors);
		static ing::regex	com_re_; /**< command matching regexp */
	};
}

#endif
