/* pipeprg.hh
* This file is part of Colorifer
* Copyright (c) 2002,2003 by Stanislav Ievlev
*
* This file is covered by the GNU General Public License,
* which should be included with colorifer as the file COPYING.
*/

#ifndef PIPE_PRG_H
#define PIPE_PRG_H

#include <sys/types.h>
#include <cerrno>

#include <assert.hh>

namespace colorifer
{
	void run(const std::string& execname, char *argv[]);

	/**
	 * This class works like popen(),
	 * run program and open two pipes to read stdout and stderror
	 */
	class pipe_prg
	{
	public:
		/**
		 * A constructor
		 * forks, replace name of the child with progname and start via execve
		 * @param args standart *argv[] structure, argv[0] - name of the program
		 * @exception Ing::glibc_error
		 */
		pipe_prg(const std::string& execname, char *args[]);
		~pipe_prg(); /**< close opened pipe descriptors */
		int	rc() const; /**< @return exit code of the child */
		int	out_fd() const; /**< @return opened file descriptor of child's stdout */
		int	err_fd() const; /**< @return opened file descriptor of child's stderr */
		/**
		 * wait until child finished it's work
		 * @exception Ing::glibc_error
		 */
		void	wait();
		/**
		 * send signal to child process
		 * @sig signal number
		 */
		void	kill(int sig);
	private:
		/**
		 * child process after fork()
		 * @param args standart *argv[] structure
		 * @exception Ing::glibc_error
		 */
		void child(const std::string& execname, char *args[]);

		int fds_stdout_[2]; /**< pipe between child's stdout and father */
		int fds_stderr_[2]; /**< pipe between child's stderr and father */

		int	child_rc_; /**< child's exit code */
		pid_t	child_pid_; /**< pid of the child process */
	};


	//multiplextor algorithm for the pipe_prg
	template <class OutF, class ErrF>
	void get_lines(const pipe_prg& prg, OutF put_out, ErrF put_err)
	{
		const int out_fd(prg.out_fd());
		const int err_fd(prg.err_fd());
		bool	has_stdout(true);
		bool	has_stderr(true);
		std::string	err_buffer, out_buffer;
		//note:
		// I have to collect string reading only one character per time to avoid race condition below:
		// std::getline reads buffer then try to read buffer again (without select) and wait for input data
		// process (e.g. gcc) try to write data but cannot 'cause colorifer already blocks this pipe
		// as a result we have a deadlock
		// I can implement string collection, reading bigger buffers, but algo will be more complicated
		while (has_stdout || has_stderr) //main multiplextor
		{
			fd_set	read_fds;
			FD_ZERO(&read_fds);

			if (has_stdout) FD_SET(out_fd, &read_fds);
			if (has_stderr) FD_SET(err_fd, &read_fds);

			char c;
			if (TEMP_FAILURE_RETRY(::select(20, &read_fds, NULL, NULL, NULL)))
				if (FD_ISSET(out_fd, &read_fds)) //this is a line from program's stdout
					if (::read(out_fd, &c, 1) > 0 )
					{
						if (c == '\n') //next string is ready
						{
							put_out(out_buffer);
							out_buffer.clear(); //we no longer need content of this buffer
						}
						else
							out_buffer += c;
					}
					else
					{ //try to process rest of line
						if (!out_buffer.empty()) put_out(out_buffer);
						has_stdout = false;
					}
				else if (FD_ISSET(err_fd, &read_fds))
					if (::read(err_fd, &c, 1) > 0 )
					{

						if (c == '\n') //next string is ready
						{
							put_err(err_buffer);
							err_buffer.clear(); //we no longer need content of this buffer
						}
						else
							err_buffer += c;
					}
					else
					{
						if (!err_buffer.empty()) put_err(err_buffer);
						has_stderr = false;
					}
				else
				{ //no more output
					has_stdout = false;
					has_stderr = false;
				}
		}
	}

}

#endif
