/* regex.hh
* This file is part of Ing library
* Copyright (c) 2002,2003 by Stanislav Ievlev
*
* This file is covered by the GNU Library General Public License,
* which should be included with libing as the file COPYING.
*/

#ifndef ING_REGEXP_H
#define ING_REGEXP_H

#include <string>
#include <vector>
#include <stdexcept>
#include <boost/ref.hpp>

namespace ing
{
	/**
	 * Generic class for process strings with regular expressions.
	 * Works with all POSIX compatible regexp implementations
	 * You must include header from implementation you want before this header
	 * @todo if we will need support for library without posix interface I can easy made this class as a template
	 */
	class regex
	{
		typedef std::vector<int> offset_vector;
	public:
		enum flags  /**< Flags determines how to compile given regular expression*/
		{
			basic = 0,                             /**<   only simple expressions */
			extended = REG_EXTENDED /**<   allow extended expressions (allow to have subexpressions) */
		};

		/**
		 * Compile given pattern into finite state machine
		 * @param pattern given regular expression
		 * @param count number of possible subexpresions
		 * @param format by default you use mordern extended format of regular expressions, you can also use old basic format
		 * @see RegularExpr::re
		 * @see Flags
		 * @exception Ing::internal_error
		 */
		regex(const std::string& pattern, int count = 10, flags format = extended):
				start_anchor(false),
				matches_(count*2)
		{
			//we are working with extented Regular Expression syntax: flag must be zero to old syntax
			if (int result = ::regcomp(&re_, pattern.c_str(), format))
			{ //compile pattern
				std::vector<char>	errbuf(::regerror(result, &re_, NULL, 0));
				::regerror(result, &re_, &errbuf[0], errbuf.size());
				throw std::runtime_error("invalid regex:" + pattern + ":" + &errbuf[0]);
			}
			if ((!pattern.empty()) && (pattern[0] == '^')) start_anchor = true; //regular expression has start anchor
		}

		/**
		* free all created resources, delete regexp's finite state machine
		* @see ing::regex::re
		*/
		~regex()
		{
			::regfree(&re_);
		}

		/**
		* checks are string matches to compiled regexp
		* @param line string to check
		* @return true if string matches to pattern, otherwise false
		*/
		bool match(const std::string& line, int offset = 0)
		{
			// we cannot match to regexp if we have start anchor
			// and see string not from the begin
			if ((offset > 0) && (start_anchor)) return false;

			//clear vector from previous running
			start_.clear();
			stop_.clear();
			std::vector<regmatch_t> matcharr(matches_);
			//run compiled expression over string
			int result = ::regexec(&re_, line.c_str() + offset, matches_, &matcharr[0], 0);
			if (!result)
			{ //fill fields
				for (int i = 0;(matcharr[i].rm_so >= 0) && (i <= matches_);++i)
				{
					start_.push_back(matcharr[i].rm_so + offset);
					stop_.push_back(matcharr[i].rm_eo + offset);
				}
			}
			return (result == 0);
		}
		/**
		* how much subexpressions was found in previous match()
		* @return number of subexpressions founded in latest checked string
		*/
		int size() const
		{
			return start_.size();
		}
		/**
		* access to subexpressions founded during latest match()
		* @return start position of the i-th founded subexpression
		*/
		int start(int i) const
		{
			return start_.at(i);
		}
		/**
		* access to subexpressions founded during latest match()
		* @return stop position of the i-th founded subexpression
		*/
		int stop(int i) const
		{
			return stop_.at(i);
		}
	private:
		//hide dangerous operations
		regex& operator=(const regex&) { return * this; }
		regex(const regex&) {}

		offset_vector start_; /**< vector of subexpression start offsets */
		offset_vector stop_; /**< vector of subexpression stop offsets */
		bool start_anchor; /**< are we have '^' anchor at the begin of the regular expression? */
		int matches_; /**< number of subexpressions */
		regex_t	re_; /**< POSIX regular expression object*/
	};

	template <class T>
	T for_each_match(ing::regex &re,const std::string& str, T f) /**< algorithm, walk throw string and call functor for each founded substring  */
	{
		int offset = 0;
		while (re.match(str, offset))
		{
			//work with main or first subexpression
			const int num = (re.size() > 1) ? 1 : 0;
			if (re.stop(num) == offset) break; //protection from infinity loop
			f(boost::cref(re.start(num)), boost::cref(re.stop(num)));
			offset = re.stop(num);
		}
		return f;
	}

	/** @return n-th substring corresponds to n-th founded subexpression */
	inline std::string substr(const ing::regex& re, int i, const std::string line)
	{
		return line.substr(re.start(i), re.stop(i) - re.start(i));
	}

	//usefull operators
	inline bool operator==(const std::string& str, ing::regex& re)
	{
		return re.match(str);
	}
	inline bool operator!=(const std::string& str, ing::regex& re)
	{
		return !re.match(str);
	}

	inline bool operator==(ing::regex& re, const std::string& str)
	{
		return re.match(str);
	}

	inline bool operator!=(ing::regex& re, const std::string& str)
	{
		return !re.match(str);
	}
}

#endif
