/* pattern.cc
* This file is part of Colorifer
* Copyright (c) 2002,2003,2004 by Stanislav Ievlev
*
* This file is covered by the GNU General Public License,
* which should be included with colorifer as the file COPYING.
*/

#include <map>
#include <stdexcept>

#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>

//for debug
#include <iostream>

#include <pattern.hh>

using namespace colorifer;

/**
 * remove static when compiller will be ready for it
 */
namespace
{
	class fg_cache: public std::map<std::string, ncurses::color_t>
	{
		typedef std::map<std::string, ncurses::color_t> map_type;
	public:
		fg_cache(): map_type()
		{
			this->insert(map_type::value_type("black", COLOR_BLACK));
			this->insert(map_type::value_type("red", COLOR_RED));
			this->insert(map_type::value_type("green", COLOR_GREEN));
			this->insert(map_type::value_type("yellow", COLOR_YELLOW));
			this->insert(map_type::value_type("blue", COLOR_BLUE));
			this->insert(map_type::value_type("magenta", COLOR_MAGENTA));
			this->insert(map_type::value_type("cyan", COLOR_CYAN));
			this->insert(map_type::value_type("white", COLOR_WHITE));
		}
	};

	class attr_cache: public std::map<std::string, ncurses::attr_t>
	{
		typedef std::map<std::string, ncurses::attr_t> map_type;
	public:
		attr_cache(): map_type()
		{
			this->insert(map_type::value_type("normal", A_NORMAL));
			this->insert(map_type::value_type("standout", A_STANDOUT));
			this->insert(map_type::value_type("underline", A_UNDERLINE));
			this->insert(map_type::value_type("reverse", A_REVERSE));
			this->insert(map_type::value_type("blink", A_BLINK));
			this->insert(map_type::value_type("dim", A_DIM));
			this->insert(map_type::value_type("bold", A_BOLD));
			this->insert(map_type::value_type("invis", A_INVIS));
			this->insert(map_type::value_type("protect", A_PROTECT));
			this->insert(map_type::value_type("altcharset", A_ALTCHARSET));
		}
	};

	static const fg_cache fg_map;
	static const attr_cache attr_map;
	
	const std::string avail_colors = "black|red|green|yellow|blue|magenta|cyan|white";
}

ing::regex pattern::com_re_("("+avail_colors+"),?([a-z]+)?[[:space:]]*");

pattern::pattern():
    fg_(end_color),
    attr_(A_NORMAL)
{}

void pattern::parse_colors(const std::string& command)
{
    if (com_re_ == command)
    {
	fg_cache::const_iterator it = fg_map.find(ing::substr(com_re_, 1, command));
	fg_ = (it == fg_map.end()) ? COLOR_WHITE : it->second;

	if(3 == com_re_.size())
	{
	    attr_cache::const_iterator it = attr_map.find(ing::substr(com_re_, 2, command));
	    attr_ = (it == attr_map.end()) ? A_NORMAL : it->second;
	}
    }
    else
	throw std::runtime_error("invalid color specification:" + command);
}


ing::regex addr_pattern::com_re_("[!]?[[:space:]]*p[[:space:]]+(.*)$");

addr_pattern::addr_pattern(const std::string& command):
    positive_(true)
{
    if (com_re_ == command)
    {
	parse_colors(ing::substr(com_re_, 1, command));
	if ('!' == command[com_re_.start(0)]) positive_ = false;
    }
    else
	throw std::runtime_error("invalid color specification:" + command);
}

void addr_pattern::fill(const std::string& line,color_map& subcolors)
{
    subcolors[0].push_back(std::make_pair(fg_, attr_));
    subcolors[line.length()].push_back(std::make_pair(end_color, A_NORMAL));
}

ing::regex regex_pattern::com_re_("/(.*?[^\\\\])/$");

regex_pattern::regex_pattern(const std::string& cmd):
	addr_pattern(cmd)
{
	const std::string command(cmd.substr(0,addr_pattern::com_re_.start(0)));
	if (com_re_ != command)
		throw std::runtime_error("invalid command:" + command);
	else
	{
		re_.reset(new ing::regex(ing::substr(com_re_, 1, command)));
		if ('!' == *command.rbegin()) positive_ = false;
	}
}

void regex_pattern::fill(const std::string& line,color_map& subcolors)
{
	if (re_->match(line) == positive_)
	    addr_pattern::fill(line,subcolors);
}

ing::regex lnum_pattern::com_re_("([[:digit:]]+)(,[+~]?|~)?([[:digit:]]+)?$");

lnum_pattern::lnum_pattern(const std::string& cmd):
	addr_pattern(cmd),
	counter_(0),
	each_(-1)
{
	const std::string command(cmd.substr(0,addr_pattern::com_re_.start(0)));
	if (com_re_ == command)
	{
	    stop_ = start_ = boost::lexical_cast<int>(ing::substr(com_re_, 1, command));
	    if (4 == com_re_.size())
	    {
		const std::string sep = ing::substr(com_re_, 2, command);
		const int num = boost::lexical_cast<int>(ing::substr(com_re_, 3, command));
		if (('~' == sep[0]) || ((sep.length()> 1) && ('~' ==sep[1])))
		{
		    each_ = num;
		}
		else
		{
		    if(sep.length()> 1 && '+' == sep[1])
			stop_ += num - 1;
		    else 
		        stop_ = num;
		}
	    }
	}
	else
	    throw std::runtime_error("invalid command format:" + command);
}

void lnum_pattern::fill(const std::string& line,color_map& subcolors)
{
    ++counter_;

    bool condition;
    if (each_ > 0)
	condition = (counter_ >= start_) && ((counter_-start_) % each_ == 0);
    else
	condition = (counter_ >= start_ && counter_ <= stop_);

    if  (condition == positive_)
	    addr_pattern::fill(line,subcolors);
}

ing::regex regsubst_pattern::com_re_("^s/(.*)/(.*)/$");

regsubst_pattern::regsubst_pattern(const std::string& cmd)
{
	if (com_re_ == cmd)
	{
	    re_.reset(new ing::regex(ing::substr(com_re_, 1, cmd)));
	    parse_colors(ing::substr(com_re_, 2, cmd));
	}
}

void regsubst_pattern::set_colors(int start, int stop, color_map& subcolors)
{
	subcolors[start].push_back(std::make_pair(fg_, attr_));
	subcolors[stop].push_back(std::make_pair(end_color, A_NORMAL));
}

void regsubst_pattern::fill(const std::string& line,color_map& subcolors)
{
	ing::for_each_match(*re_, line,
			    boost::bind(&regsubst_pattern::set_colors, this,
					_1, _2, boost::ref(subcolors)));
}
